/* File:      chain.C
   Date:      5/05/03
   Author:    C. Paciorek
   Purpose:   basic code for running MCMC - initialization, iterations, reporting, and storage
   Compile:   g++ -c  -Wall -ansi
   Usage:     N/A
   Arguments: N/A
   Requires:  see just below
   Reference: all coded by C. Paciorek
*/

#include "base.h"
#include "chain.h"
#include "cov_mat.h"
#include "parm.h"
#include "dataset.h"
#include "transform.h"
#include "covariates.h"
#include "ran1.h"

// see chain.h for comments on what the objects and methods are

chain::chain(string log_fn_in,string eval_fn_in,string fun_form_fn_in,string dat_fn_in,string init_fn_in,string final_fn_in,string parm_fn_in,string out_fn_in,int num_its_in,int storage_interval_in,int storage_space_in){

  int j,p,q;
  int repeat;

  //////////////////////////////////////
  // Parameters and initial values
  //////////////////////////////////////

  double df_parm1, df_parm2;
  
  double fun_mu_init,fun_logsigma_init,fun_mu_mean,fun_mu_sd,fun_logsigma_mean,fun_logsigma_sd,fun_omega_propsd,fun_mu_propsd,fun_logsigma_propsd;

  double logeta_init, logeta_mean, logeta_sd, logeta_propsd;

  double fun_nu_init,fun_nu_parm1,fun_nu_parm2, fun_nu_propsd;  // Matern differentiability parameter for function; not used when have stationary function prior (instead, kernel parameters are used)

  // initial and prior values for (NSGP) kernel correlation or (SGP) stationary function correlation
  double kernel_nu_init;
  double* kernel_logkappa_init;
  double* kernel_gamma_init;
  double kernel_logkappa_parm1;
  double kernel_logkappa_parm2;
  double* kernel_logkappa_propsd;
  double* kernel_gamma_propsd;
  double kernel_nu_parm1;  
  double kernel_nu_parm2;  
  double kernel_nu_propsd;  

  //initial and prior values for nonstationary eigenstructure
  double* eigenval_mu_init;
  double* eigenval_logsigma_init;

  double* eigenvec_mu_init;
  double* eigenvec_logsigma_init;

  double eigenval_mu_mean;
  double eigenval_mu_sd;
  double* eigenval_mu_propsd;

  double eigenvec_mu_mean;
  double eigenvec_mu_sd;
  double* eigenvec_mu_propsd;

  double eigenval_logsigma_mean;
  double eigenval_logsigma_sd;
  double* eigenval_logsigma_propsd;

  double eigenvec_logsigma_mean;
  double eigenvec_logsigma_sd;
  double* eigenvec_logsigma_propsd;

  double* eigenval_omega_propsd;
  double* eigenvec_omega_propsd;

  // number of covariates and eigenelements 
  num_datapoints=covariates->num_datapoints;  // train plus test points 
  num_testpoints=covariates->num_testpoints;
  num_trainpoints=covariates->num_trainpoints;
  num_covars=covariates->num_covars;
  num_angles=covariates->num_angles;
  num_eigenvecs=covariates->num_eigenvecs;

  // filenames
  final_fn=final_fn_in;
  log_fn=log_fn_in;
  eval_fn=eval_fn_in;
  out_fn=out_fn_in;

  num_its=num_its_in;

  storage_interval=storage_interval_in;
  storage_space=storage_space_in;
  it_run=0;

  block=1;
  
  //////////////////////////////////////////////////
  // create storage space for parameter iterates
  //////////////////////////////////////////////////

  if(storage_space){  // don't store anything if 0 specified
    // loglik, logprior, fun hyperparameters, and eta storage, plus MSE running values
    storage1=new mat_pp<double>(storage_space,9);

    // function iterates
    storage2=new mat_pp<double>(storage_space,num_datapoints);

    // kernel parameters
    storage3=new mat_pp<double>(storage_space,num_covars+num_angles+1);

    if(prior==NON_STATIONARY){
      // eigenstructure storage

      // eigenvalues at the training covariates
      storage4=new mat_pp<double>(storage_space,num_covars*num_trainpoints);
      
      // mu and sigma values for the eigenvalue processes
      storage5=new mat_pp<double>(storage_space,num_covars*2);

      if(num_covars>1){
        // eigenvector coordinates at the training covariates
        storage6=new mat_pp<double>(storage_space,num_eigenvecs*num_trainpoints);
      }
    }
  }


  ////////////////////////////////
  // Parameter initialization
  ////////////////////////////////

  char text[5000];    // just a buffer for unread text - make this larger if comments in the init or parm files are very long
  ifstream init(init_fn_in.c_str());


  // function initialization
  fun_omega_init=new double[num_trainpoints];
  init >> repeat;  
  if(repeat==OMEGA_REP){
    init >> fun_omega_init[0];
    for(j=1;j<num_trainpoints;j++){ 
      fun_omega_init[j]=fun_omega_init[0];  
    }
  }
  else{
    for(j=0;j<num_trainpoints;j++){ 
      init >> fun_omega_init[j];  
    }
  }
  init.getline(text, sizeof(text));

  // function hyperparameter initialization
  init >> fun_mu_init; 
  init.getline(text, sizeof(text));
  init >> fun_logsigma_init;
  init.getline(text, sizeof(text));
  if(prior==NON_STATIONARY){
    init >> fun_nu_init;
    init.getline(text, sizeof(text));
  }
  init >> logeta_init;
  init.getline(text, sizeof(text));

  // function correlation structure initialization
  // if stationary GP used, then these are the hyperparameters determining the correlation structure; if NS GP used, these are the hyperparameters of the shared kernel structure
  kernel_logkappa_init=new double[num_covars];
  for(p=0;p<num_covars;p++){
    init >> kernel_logkappa_init[p];
  }
  init.getline(text, sizeof(text));
  if(num_angles){
    kernel_gamma_init=new double[num_angles];
    for(q=0;q<num_angles;q++){
      init >> kernel_gamma_init[q];
    }
    init.getline(text, sizeof(text));
  }
  init >> kernel_nu_init;
  init.getline(text, sizeof(text));

  if(prior==NON_STATIONARY){
  // eigenval initialization
    eigenval_omega_init=new (double*)[num_covars];
    eigenval_mu_init=new double[num_covars];
    eigenval_logsigma_init=new double[num_covars];
    for(p=0;p<num_covars;p++){
      eigenval_omega_init[p]=new double[num_trainpoints];
      init >> repeat;
      if(repeat==OMEGA_REP){
	init >> eigenval_omega_init[p][0];
	for(j=1;j<num_trainpoints;j++){ 
	  eigenval_omega_init[p][j]=eigenval_omega_init[p][0];  
	}
      }
      else{
	for(j=0;j<num_trainpoints;j++){ 
	  init >> eigenval_omega_init[p][j];  
	}
      }
      init.getline(text, sizeof(text));

      init >> eigenval_mu_init[p]; 
      init.getline(text, sizeof(text));
      init >> eigenval_logsigma_init[p];
      init.getline(text, sizeof(text));
    }
    if(num_covars>1){
      // eigenvector coordinate initialization
      eigenvec_omega_init=new (double*)[num_eigenvecs];
      eigenvec_mu_init=new double[num_eigenvecs];
      eigenvec_logsigma_init=new double[num_eigenvecs];
      for(p=0;p<num_eigenvecs;p++){
	eigenvec_omega_init[p]=new double[num_trainpoints];
	// if repeat==1 then initialize all same; otherwise take values from file
	init >> repeat;
	if(repeat==OMEGA_REP){
	  init >> eigenvec_omega_init[p][0];
	  for(j=1;j<num_trainpoints;j++){ 
	    eigenvec_omega_init[p][j]=eigenvec_omega_init[p][0];  
	  }
	}
	else{
	  for(j=0;j<num_trainpoints;j++){ 
	    init >> eigenvec_omega_init[p][j];  
	  }
	}
        init.getline(text, sizeof(text));

	init >> eigenvec_mu_init[p]; 
        init.getline(text, sizeof(text));
	init >> eigenvec_logsigma_init[p];
        init.getline(text, sizeof(text));
      }
    }
  }


  ///////////////////////////////////
  /// random seed stuff
  ///////////////////////////////////

  // initialize random variable
  idum= new int;
  init >> *idum;
  init.getline(text,sizeof(text));

  // fill ran1 stored values if not yet initialized
  if(*idum<=0){
    ran1(idum,GENERATE,"a");
  }
  else{
    ran1(idum,READ,text);
  }

  /////////////////////////////////////
  // Fixed hyperparameter values
  /////////////////////////////////////

  ifstream parm(parm_fn_in.c_str());
  
  // function hyperparameters

  // degrees of freedom prior
  parm >> df_parm1;
  parm >> df_parm2;
  parm.getline(text,sizeof(text));
 
  parm >> fun_mu_mean;
  parm >> fun_mu_sd;
  parm.getline(text,sizeof(text));
  parm >> fun_logsigma_mean;
  parm >> fun_logsigma_sd;
  parm.getline(text,sizeof(text));
  if(prior==NON_STATIONARY){
    parm >> fun_nu_parm1;
    parm >> fun_nu_parm2;
    parm.getline(text,sizeof(text));
  }
  parm >> fun_omega_propsd;
  parm.getline(text,sizeof(text));
  parm >> fun_mu_propsd;
  parm.getline(text,sizeof(text));
  parm >> fun_logsigma_propsd;
  parm.getline(text,sizeof(text));
  
  if(prior==NON_STATIONARY){
    parm >> fun_nu_propsd;
    parm.getline(text,sizeof(text));
  }
  
  parm >> logeta_mean;
  parm >> logeta_sd;
  parm.getline(text,sizeof(text));
  parm >> logeta_propsd;
  parm.getline(text,sizeof(text));
  
  // function correlation structure parameters
  // if stationary GP used, then these are the hyperparameters determining the correlation structure; if NS GP used, these are the hyperparameters of the shared kernel structure

  parm >> kernel_logkappa_parm1;
  parm >> kernel_logkappa_parm2;
  parm.getline(text,sizeof(text));
  parm >> kernel_nu_parm1;
  parm >> kernel_nu_parm2;
  parm.getline(text,sizeof(text));
  kernel_logkappa_propsd=new double[num_covars];
  for(p=0;p<num_covars;p++){
    parm >> kernel_logkappa_propsd[p];
  }
  parm.getline(text,sizeof(text));
  if(num_angles){
    kernel_gamma_propsd=new double[num_angles];
    for(p=0;p<num_angles;p++){   
      parm >> kernel_gamma_propsd[p];
    }
    parm.getline(text,sizeof(text));
  }
  parm >> kernel_nu_propsd;
  parm.getline(text,sizeof(text));
    
  if(prior==NON_STATIONARY){
    parm >> eigenval_mu_mean;
    parm >> eigenval_mu_sd;
    parm.getline(text,sizeof(text));
    parm >> eigenval_logsigma_mean;
    parm >> eigenval_logsigma_sd;
    parm.getline(text,sizeof(text));
   
    eigenval_omega_propsd=new double[num_covars];
    eigenval_mu_propsd=new double[num_covars];
    eigenval_logsigma_propsd=new double[num_covars];
    for(p=0;p<num_covars;p++){
      parm >> eigenval_omega_propsd[p];
      parm.getline(text,sizeof(text));
      parm >> eigenval_mu_propsd[p];
      parm.getline(text,sizeof(text));
      parm >> eigenval_logsigma_propsd[p];
      parm.getline(text,sizeof(text));
    }
    if(num_covars>1){
      // eigenvector coordinate parameters
      parm >> eigenvec_mu_mean;
      parm >> eigenvec_mu_sd;
      parm.getline(text,sizeof(text));
      parm >> eigenvec_logsigma_mean;
      parm >> eigenvec_logsigma_sd;
      parm.getline(text,sizeof(text));
      eigenvec_omega_propsd=new double[num_eigenvecs];
      eigenvec_mu_propsd=new double[num_eigenvecs];
      eigenvec_logsigma_propsd=new double[num_eigenvecs];
      for(p=0;p<num_eigenvecs;p++){
        parm >> eigenvec_omega_propsd[p];
        parm.getline(text,sizeof(text));
        parm >> eigenvec_mu_propsd[p];
        parm.getline(text,sizeof(text));
        parm >> eigenvec_logsigma_propsd[p];
        parm.getline(text,sizeof(text)); 
     }
    }
  }
  // I could possibly order the parameter initialization more rationally so I don't need as many function calls to add initialization information, but not sure

  ////////////////////////////////////////////////////////////////////
  // initialize parameters and structures
  /////////////////////////////////////////////////////////////////////

  // log of noise sd
  eta = new nonspatial_parm(ETA,logeta_init,logeta_mean,logeta_sd,logeta_propsd,&log_transform);
  // holds data, has functions dealing with likelihood and evaluation criteria
  data=new dataset(dat_fn_in,fun_form_fn_in,&(eta->psi->value),&(eta->psi->value_stored),df_parm1,df_parm2);

  // holds correlation structure of function
  kernel=new spatial_parm(KERNEL,
                                 kernel_logkappa_init,
                                 kernel_gamma_init,
                                 kernel_nu_init,
                                 kernel_logkappa_parm1,
                                 kernel_logkappa_parm2,
                                 kernel_nu_parm1,
                                 kernel_nu_parm2,
                                 kernel_logkappa_propsd,
                                 kernel_gamma_propsd,
                                 kernel_nu_propsd);

  if(prior==STATIONARY){
    fun = new spatial_parm(FUN,  
                                  fun_omega_init,
                                  fun_mu_init,
                                  fun_logsigma_init,
                                  0.0,         // for stationary case, nu is in kernel structure
                                  fun_mu_mean,
                                  fun_mu_sd,
                                  fun_logsigma_mean,
                                  fun_logsigma_sd,
                                  0.0,       // for stationary case, nu is in kernel structure
                                  0.0,       // for stationary case, nu is in kernel structure
                                  fun_omega_propsd,
                                  fun_mu_propsd,
                                  fun_logsigma_propsd,
                                  0.0,       // for stationary case, nu is in kernel structure
                                  &identity);    

    // most of code below exchanges information between the various data structures
    
    fun->add_child(kernel->stat_cov,kernel->stat_cov_full);   // associate function prior covariance with fun data object; stat_cov_full is the covariance of train and test, stat_cov is just the training set
    kernel->add_child(fun);   // associate the function with the covariance structure
    data->add_info(fun->psi->value,fun->psi_full,kernel->stat_cov->lnz_R,&(fun->sigma->value));   // pass function and covariance pointers to data object
    fun->ksi_propose(TRUE);  // calculates value of chi vector; needs to be done after Cholesky of sum of data and prior covariance is done in data->add_info

    // nonstationary covariance objects not used
    cov=0;
    cov_full=0;
  }
  else{
    // vector of eigenvalue processes
    eigenval=new (spatial_parm*)[num_covars];

    // pointers to vectors of eigenvalue process values
    eigenval_dbl=new (double*)[num_covars];         // train set
    eigenval_full_dbl=new (double*)[num_covars];    // train plus test set
    
    for(p=0;p<num_covars;p++){
      eigenval[p]=new spatial_parm(FUN_COVAR_EIGVAL,  // 0.0 values are because nu for the process is associated with the kernel object
                                          eigenval_omega_init[p],   
                                          eigenval_mu_init[p],
                                          eigenval_logsigma_init[p],
                                          0.0,
                                          eigenval_mu_mean,
                                          eigenval_mu_sd,
                                          eigenval_logsigma_mean,
                                          eigenval_logsigma_sd,
                                          0.0,
                                          0.0,
                                          eigenval_omega_propsd[p],
                                          eigenval_mu_propsd[p],
                                          eigenval_logsigma_propsd[p],
                                          0.0,
                                          &log_transform);
    }
    if(num_covars>1){   // if so, there are eigenvectors
      // vector of eigenvector processes
      eigenvec=new (spatial_parm*)[num_eigenvecs];

      // pointers to vectors of eigenvector process values
      eigenvec_dbl=new (double*)[num_eigenvecs];       // train set
      eigenvec_full_dbl=new (double*)[num_eigenvecs];  // train plus test sets

      for(p=0;p<num_eigenvecs;p++){
        eigenvec[p]=new spatial_parm(FUN_COVAR_EIGVEC,      // 0.0 values are because nu for the process is associated with the kernel object
                                            eigenvec_omega_init[p],
                                            eigenvec_mu_init[p],
                                            eigenvec_logsigma_init[p],
                                            0.0,
                                            eigenvec_mu_mean,
                                            eigenvec_mu_sd,
                                            eigenvec_logsigma_mean,
                                            eigenvec_logsigma_sd,
                                            0.0,
                                            0.0,
                                            eigenvec_omega_propsd[p],
                                            eigenvec_mu_propsd[p],
                                            eigenvec_logsigma_propsd[p],
                                            0.0,
                                            &identity);
      }
    }
    
    fun = new spatial_parm(FUN,                     
                                  fun_omega_init,
                                  fun_mu_init,
                                  fun_logsigma_init,
                                  fun_nu_init,
                                  fun_mu_mean,
                                  fun_mu_sd,
                                  fun_logsigma_mean,
                                  fun_logsigma_sd,
                                  fun_nu_parm1,
                                  fun_nu_parm2,
                                  fun_omega_propsd,
                                  fun_mu_propsd,
                                  fun_logsigma_propsd,
                                  fun_nu_propsd,
                                  &identity);   
    
    // exchange information between structures

    kernel->add_child(fun);   // associate fun with kernel object
    for(p=0;p<num_covars;p++){
      // associate eigenprocess correlation structure with eigenvalue processes
      eigenval[p]->add_child(kernel->stat_cov,kernel->stat_cov_full);
      eigenval[p]->add_child(fun);

      eigenval_dbl[p]=eigenval[p]->psi->value;
      eigenval_full_dbl[p]=eigenval[p]->psi_full;        
      eigenval[p]->sample_full();     // initial sample (train plus test) of eigenvalue processes values
    }
    if(num_covars>1){
      // associate eigenprocesses with eigenprocess correlation structure
      kernel->add_child(eigenval,eigenvec);
      for(p=0;p<num_eigenvecs;p++){
        eigenvec[p]->add_child(kernel->stat_cov,kernel->stat_cov_full);
        eigenvec[p]->add_child(fun);
        eigenvec_dbl[p]=eigenvec[p]->psi->value;
        eigenvec_full_dbl[p]=eigenvec[p]->psi_full;
        eigenvec[p]->sample_full();     // initial sample (train plus test) of eigenvalue processes values
      }
    }
    else{
      kernel->add_child(eigenval,0); // associate eigenprocesses with eigenprocess correlation structure
    }
    // nonstationary correlation structure for training set
    cov=new sparse_cov_mat_nonstat(num_trainpoints,num_covars,eigenval_dbl,eigenvec_dbl,&(fun->nu->value));
    // nonstationary correlation structure for test set
    cov_full=new sparse_cov_mat_nonstat(num_datapoints,num_covars,eigenval_full_dbl,eigenvec_full_dbl,&(fun->nu->value));
    fun->add_child(cov,cov_full);  // associate covariance with fun object
    data->add_info(fun->psi->value,fun->psi_full,cov->lnz_R,&(fun->sigma->value));      
    fun->ksi_propose(TRUE);        // get initial values of chi vector (based on Cholesky of sum of data and prior covariance done in data->add_info()
    
    // Cleanup

    delete [] eigenval_mu_init;
    delete [] eigenval_logsigma_init;
    delete [] eigenval_omega_propsd;
    delete [] eigenval_mu_propsd;
    delete [] eigenval_logsigma_propsd;
    if(num_covars>1){
      delete [] eigenvec_mu_init;
      delete [] eigenvec_logsigma_init;
      delete [] eigenvec_omega_propsd;
      delete [] eigenvec_mu_propsd;
      delete [] eigenvec_logsigma_propsd;
    }
  }
  
  delete [] kernel_logkappa_propsd;
  delete [] kernel_logkappa_init;
  if(num_angles){
    delete [] kernel_gamma_propsd;
    delete [] kernel_gamma_init;
  }
}


chain::~chain(void){
  int p;
  if(storage_space){
    delete storage1;
    delete storage2;
    delete storage3;
    if(prior==NON_STATIONARY){
      delete storage4;
      if(num_covars>1){
        delete storage5;
        delete storage6;
      }
    }
  }
  delete fun; 
  delete eta;
  delete data;

  delete kernel;
  // delete [] fun_omega_init;  // deleted in hyperparm_vec
  if(prior==NON_STATIONARY){
    delete [] eigenval_dbl;
    delete [] eigenval_full_dbl;
    for(p=0;p<num_covars;p++){
      delete eigenval[p];
    }
    delete [] eigenval;
    delete [] eigenval_omega_init;    /*  I delete the component vectors in hyperparm_vec since omega_init sent there */

    if(num_covars>1){
      delete [] eigenvec_dbl;
      delete [] eigenvec_full_dbl;
      for(p=0;p<num_eigenvecs;p++){
	delete eigenvec[p];
      }
      delete [] eigenvec;
      delete [] eigenvec_omega_init;    /*  I delete the component vectors in hyperparm_vec since omega_init sent there */

    }
    delete cov;
    delete cov_full;
  }
}

void chain::do_a_step(int it_cur,int burnin){
  // does a single round of parameter updating
  int accepted=0;
  int p;

  int mode=POSTMEAN;  // alternative is SIMPLE, which does the centered-joint proposal (which is the same as the centered proposal
  // propose only kappa, gamma, nu
  if(prior==STATIONARY){  
    // propose only mu, sigma, omega
    fun->step(mode,NON_LANGEVIN,FIXED,FIXED,NOT_FIXED,NOT_FIXED,FIXED,NOT_FIXED);
    accepted=eta->step(NOT_FIXED);
    if(accepted && mode==POSTMEAN){
      // update fun->ksi if eta accepted since ksi changes
      fun->ksi_propose(TRUE);
    }
    // propose kappa, gamma, nu; nu is the nu for the function
    kernel->step(mode,NON_LANGEVIN,NOT_FIXED,NOT_FIXED,FIXED,FIXED,NOT_FIXED,FIXED);
  }
  else{
    // propose only mu, sigma, nu, omega
    fun->step(mode,NON_LANGEVIN,FIXED,FIXED,NOT_FIXED,NOT_FIXED,NOT_FIXED,NOT_FIXED); 
    accepted=eta->step(NOT_FIXED);
    if(accepted && mode==POSTMEAN){
      // update fun->ksi if eta accepted since ksi changes
      fun->ksi_propose(TRUE);
    }   
    // propose kappa, gamma; nu fixed for covariance processes
    kernel->step(mode,NON_LANGEVIN,NOT_FIXED,NOT_FIXED,FIXED,FIXED,FIXED,FIXED);

  }
  if(prior==NON_STATIONARY){
    for(p=0;p<num_covars;p++){
      // propose only mu,omega
      eigenval[p]->step(mode,NON_LANGEVIN,FIXED,FIXED,NOT_FIXED,FIXED,FIXED,NOT_FIXED);
    }
    if(num_covars>1){
      for(p=0;p<num_eigenvecs;p++){
        // propose only omega
        eigenvec[p]->step(mode,NON_LANGEVIN,FIXED,FIXED,FIXED,FIXED,FIXED,NOT_FIXED); 
      }
    }
  }
  
  // get sample of f on test data
  if(num_testpoints && !burnin){ // && it_cur%5==0){
    if(prior==NON_STATIONARY){
      // find full covariance and Cholesky of kernel processes for train and test points
      kernel->stat_cov_full->fill(NOT_FIRST);
      kernel->stat_cov_full->gen_chol(NO_WARN);
      for(p=0;p<num_covars;p++){
        // get sample of eigenvalues for train and test points
        eigenval[p]->sample_full();
      }
      if(num_covars>1){
        for(p=0;p<num_eigenvecs;p++){
          eigenvec[p]->sample_full();  // get sample of eigenvectors for train and test points
        }
      }
    }
    fun->sample_full();  // calculate sample function; covariance and Cholesky for fun done here
  }
  if(!burnin){
    data->update_postmean();
  }
}

void chain::store(int it_cur,int store,int burnin){
  // stores parm values from current iteration
  double tmp, tmp2; int j,p;
  int num=100;
  double logprior;
  
  logprior=fun->mu->logprior+fun->sigma->logprior+fun->omega->logprior+eta->psi->logprior+data->logprior_df;
  for(p=0;p<num_covars;p++){
    logprior+=kernel->kappa[p]->logprior;
  }
  for(p=0;p<num_angles;p++){
    logprior+=kernel->gamma[p]->logprior;
  }
  logprior+=kernel->nu->logprior;
  if(prior==NON_STATIONARY){
    logprior+=fun->nu->logprior;
    for(p=0;p<num_covars;p++){
      logprior+=eigenval[p]->mu->logprior+eigenval[p]->sigma->logprior+eigenval[p]->omega->logprior;
    }
    if(num_covars>1){
      for(p=0;p<num_eigenvecs;p++){
        logprior+=eigenvec[p]->mu->logprior+eigenvec[p]->sigma->logprior+eigenvec[p]->omega->logprior;
      }
    }
  }


  if(it_cur%1000==0 && it_cur>0 && !burnin){
    ofstream evalfile;
    evalfile.open(eval_fn.c_str(),ios::app);

    data->finalize_postmean();  // calculates current posterior mean function
    double* tmp;
    tmp=data->calculate_MSE();  // train MSE and FVU; test MSE and FVU
    evalfile << it_cur << "\t" << tmp[0] << "\t" << tmp[1] << "\t" << tmp[2] << "\t" << tmp[3];
    delete [] tmp;
    evalfile.close();
  }

  if(it_cur%num==0 && it_cur>0){
    // print out acceptance rates and current values
    ofstream logfile;
    if(it_cur==num){
      logfile.open(log_fn.c_str());
    }
    else{
      logfile.open(log_fn.c_str(),ios::app);
    }
    logfile << "Finished " << it_cur << " iterations.\n";
    // Langevin stuff -- tell how many times truncation of gradient used
    if(fun->oldgrad_exceeded+fun->newgrad_exceeded){
      logfile << "Gradient truncated: " << fun->oldgrad_exceeded << " " << fun->newgrad_exceeded << endl;
      fun->oldgrad_exceeded=0;
      fun->newgrad_exceeded=0;
    }

    // calculate and output acceptance rates    
    logfile << setprecision(6) << "Current acceptance rates: " << data->df << " " << data->loglik << " " << logprior <<"\n";

    // function omega
    tmp=(double) fun->num_accept_omega/it_cur;
    tmp2=(double) fun->run_num_accept_omega/num;
    fun->run_num_accept_omega=0;
    logfile << setprecision(2) << "M,O: " << tmp << "\t" << tmp2 << endl;
    // function mu
    tmp=(double)fun->num_accept_mu/it_cur;
    tmp2=(double) fun->run_num_accept_mu/num;
    fun->run_num_accept_mu=0;
    logfile << setprecision(2) <<"M,M: " << tmp << "\t" << tmp2 << "\t" << fun->mu->value << endl;
    // function sigma
    tmp=(double)fun->num_accept_sigma/it_cur;
    tmp2=(double) fun->run_num_accept_sigma/num;
    fun->run_num_accept_sigma=0;
    logfile << setprecision(2) <<"M,S: " << tmp << "\t" << tmp2 << "\t" << fun->sigma->value << endl;
    if(prior==NON_STATIONARY){
      // function nu
      tmp=(double)fun->num_accept_nu/it_cur;
      tmp2=(double) fun->run_num_accept_nu/num;
      fun->run_num_accept_nu=0;
      logfile << setprecision(2) <<"M,V: " << tmp << "\t" << tmp2 << "\t" << fun->nu->value << endl;
    }
    // eta (noise variance)
    tmp=(double) eta->num_accept/it_cur;
    tmp2=(double) eta->run_num_accept/num;
    eta->run_num_accept=0;
    logfile << setprecision(2) << "E:   " << tmp << "\t" << tmp2 << "\t" << eta->psi->value << endl;

    // kernel nu
    tmp=(double)kernel->num_accept_nu/it_cur;
    tmp2=(double) kernel->run_num_accept_nu/num;
    kernel->run_num_accept_nu=0;
    logfile << setprecision(2) <<"K,V: " << tmp << "\t" << tmp2 << "\t" << kernel->nu->value << endl;
    // kernel kappa values
    for(p=0;p<num_covars;p++){
      tmp=(double) kernel->num_accept_kappa[p]/it_cur;
      tmp2=(double) kernel->run_num_accept_kappa[p]/num;
      kernel->run_num_accept_kappa[p]=0;
      logfile << setprecision(2) << "K,K," << p << ": "  << tmp << "\t" << tmp2 << "\t" << kernel->kappa[p]->value <<  endl;
    }
    // kernel gamma values
    for(p=0;p<num_angles;p++){
      tmp=(double) kernel->num_accept_gamma[p]/it_cur;
      tmp2=(double) kernel->run_num_accept_gamma[p]/num;
      kernel->run_num_accept_gamma[p]=0;
      logfile << setprecision(2) << "K,G," << p << ": "  << tmp << "\t" << tmp2 << "\t" << kernel->gamma[p]->value <<  endl;
    }
    if(prior==NON_STATIONARY){
      for(p=0;p<num_covars;p++){
        // eigenval omega values
	tmp=(double)eigenval[p]->num_accept_omega/it_cur;
	tmp2=(double) eigenval[p]->run_num_accept_omega/num;
	eigenval[p]->run_num_accept_omega=0;
	logfile << setprecision(2) <<"L,O," << p << ": " << tmp << "\t" << tmp2 << endl;
        // eigenval mu values
	tmp=(double)eigenval[p]->num_accept_mu/it_cur;
	tmp2=(double) eigenval[p]->run_num_accept_mu/num;
	eigenval[p]->run_num_accept_mu=0;
	logfile << setprecision(2) << "L,M," << p << ": " << tmp << "\t" << tmp2 << "\t" << eigenval[p]->mu->value << endl;
        // eigenval sigma values
	tmp=(double)eigenval[p]->num_accept_sigma/it_cur;
	tmp2=(double) eigenval[p]->run_num_accept_sigma/num;
	eigenval[p]->run_num_accept_sigma=0;
	logfile << setprecision(2) << "L,S," << p << ": " << tmp << "\t" << tmp2 << "\t" << eigenval[p]->sigma->value << endl;
      }
      if(num_covars>1){
	for(p=0;p<num_eigenvecs;p++){
          // eigenvector omega values
	  tmp=(double)eigenvec[p]->num_accept_omega/it_cur;
	  tmp2=(double) eigenvec[p]->run_num_accept_omega/num;
	  eigenvec[p]->run_num_accept_omega=0;
	  logfile << setprecision(2) <<"G,O," << p << ": " << tmp << "\t" << tmp2 << endl;
	}
      }
    }
    logfile.close();
  }

  // save iterates to storage (every storage_interval iterations)
  if(store && it_cur%storage_interval==0){
    storage1->vals[it_run][0]=data->loglik;
    storage1->vals[it_run][1]=logprior;
    storage1->vals[it_run][2]=fun->mu->value;
    storage1->vals[it_run][3]=fun->sigma->value;
    if(prior==NON_STATIONARY){
      storage1->vals[it_run][4]=fun->nu->value;
    }
    else{
      storage1->vals[it_run][4]=kernel->nu->value;
    }
    storage1->vals[it_run][5]=eta->psi->value;
    storage1->vals[it_run][6]=data->df;
    
    data->current_eval();
    storage1->vals[it_run][7]=data->current_MSE_train;
    if(covariates->num_testpoints){
      storage1->vals[it_run][8]=data->current_MSE_test;
    }
    else{
      storage1->vals[it_run][8]=0.0;
    }
    for(j=0;j<num_trainpoints;j++){
      storage2->vals[it_run][j]=fun->psi->value[j];
    }
    for(j=num_trainpoints;j<num_datapoints;j++){
      storage2->vals[it_run][j]=fun->psi_full[j];
    }

    for(p=0;p<num_covars;p++){
      storage3->vals[it_run][p]=kernel->kappa[p]->value;
    }
    for(p=0;p<num_angles;p++){
      storage3->vals[it_run][p+num_covars]=kernel->gamma[p]->value;
    }
    storage3->vals[it_run][num_covars+num_angles]=kernel->nu->value;
  
    if(prior==NON_STATIONARY){
      for(p=0;p<num_covars;p++){
        // eigenvals
        storage5->vals[it_run][p*2]=eigenval[p]->mu->value;
        storage5->vals[it_run][1+p*2]=eigenval[p]->sigma->value;
        for(j=0;j<num_trainpoints;j++){
          storage4->vals[it_run][j+p*(num_trainpoints)]=eigenval[p]->psi->value[j];
        }
      }
      for(p=0;p<num_eigenvecs;p++){
        // eigenvecs
        if(num_covars>1){
          for(j=0;j<num_trainpoints;j++){
            storage6->vals[it_run][j+p*(num_trainpoints)]=eigenvec[p]->psi->value[j];
          }
        }
      }
    }
    it_run++;
  }
  if(store && it_run>0 && it_run%storage_space==0){
    // print out stored values
    report(it_cur,CURRENT);
  }
}

void chain::report(int it_cur,int status){
  // print out stored values
  int j,p;
  cout << "Process: wait. Writing output or final values at iteration " << it_cur << " to files " << out_fn << "; ";
  time_t t1=time(0);
  char* timeout=ctime(&t1);
  cout << timeout << endl; cout.flush();

  if(it_run){
    it_run=0;
    cout << "Writing output at iteration " << it_cur << " to file\n";
    
    string out_fn_tmp;
    ofstream output;
    
    //string blocks=itos(block);
    
    
    string blocks="";  // I'm appending to the output file, formerly blocks told which file to print to
    out_fn_tmp=out_fn+"1" + blocks;
    output.open(out_fn_tmp.c_str(),ios::app);
    //output.open(out_fn_tmp.c_str());
    output << storage1;
    output.close();

    
    out_fn_tmp=out_fn+"2" + blocks;
    //output.open(out_fn_tmp.c_str());
    output.open(out_fn_tmp.c_str(),ios::app);
    output << storage2;
    output.close();
    
    out_fn_tmp=out_fn+"3" + blocks;
    //output.open(out_fn_tmp.c_str());
    output.open(out_fn_tmp.c_str(),ios::app);
    output << storage3;
    output.close();

    if(prior==NON_STATIONARY){
      out_fn_tmp=out_fn+"4" + blocks;
      //output.open(out_fn_tmp.c_str());
      output.open(out_fn_tmp.c_str(),ios::app);
      output << storage4;
      output.close();

      out_fn_tmp=out_fn+"5" + blocks;
      output.open(out_fn_tmp.c_str(),ios::app);
      //output.open(out_fn_tmp.c_str());
      output << storage5;
      output.close();

      if(num_covars>1){
	out_fn_tmp=out_fn+"6" + blocks;
	//output.open(out_fn_tmp.c_str());
	output.open(out_fn_tmp.c_str(),ios::app);
	output << storage6;
	output.close();
      }
    }


    block++;
  }

  // storage and output both contain permuted versions -- unpermute when analyzing as necessary
  
  // write final values to file
  ofstream final(final_fn.c_str());
  final << setprecision(15);
  
  final << OMEGA_NOREP << " ";       // no repeat initialization
  for(j=0;j<num_trainpoints;j++){
    final << fun->omega->value[j] << " ";
  }
  final << endl << fun->mu->base_value << endl; 
  final << fun->sigma->base_value << endl;
  if(prior==NON_STATIONARY){
    final << fun->nu->base_value << endl;
  }
  final << eta->psi->base_value << endl;
  
  for(p=0;p<num_covars;p++){
    final << kernel->kappa[p]->base_value << " ";
  }
  final << endl;
  for(p=0;p<num_angles;p++){
    final << kernel->gamma[p]->base_value << " ";
  }
  final << endl;
  final << kernel->nu->base_value << endl;

  if(prior==NON_STATIONARY){
    // eigenval output
    for(p=0;p<num_covars;p++){
      final << OMEGA_NOREP << " ";       // no repeat initialization
      for(j=0;j<num_trainpoints;j++){ 
        final <<  eigenval[p]->omega->value[j] << " ";
      }
      final << endl <<  eigenval[p]->mu->base_value << endl; 
      final << eigenval[p]->sigma->base_value << endl;
    }
    // eigenvecs output
    if(num_covars>1){
      for(p=0;p<num_eigenvecs;p++){
        final << OMEGA_NOREP << " ";       // no repeat initialization
        for(j=0;j<num_trainpoints;j++){ 
          final <<  eigenvec[p]->omega->value[j] << " ";
        }
        final << endl <<  eigenvec[p]->mu->base_value << endl; 
        final << eigenvec[p]->sigma->base_value << endl;
      }
    }
  }
  char line[1024];
  for(int j=0;j<1024;j++){
    line[j]=' ';
  }
  // write out current random seed so can pick up where left off
  ran1(idum,WRITE,line);
  final << line << endl;

  final.close();
  cout << "Process: done writing.\n"; cout.flush();
}
























